import argparse
import os
import json
import pathlib
from typing import Optional
from ppline.easy.command import Run



def _arg_directory(arg):
    expanded = os.path.expanduser(arg)
    if os.path.isdir(expanded):
        return os.path.abspath(os.path.normpath(expanded))
    else:
        raise Exception(f'Argument {arg} must be an existing directory.')


def _arg_pipeline_config(arg):
    # if arg is None:
    #     print('no arg for pipeline')
        # raise TypeError(f'Must specify .yml pipeline config file.')
    if pathlib.Path(arg).suffix not in ['.yml', '.yaml']:
        raise Exception(f'Argument {arg} must be a .yml/.yaml file.')
    if os.path.exists(arg)==False:
        raise Exception(f'Cannot find {arg} at your directory')
    return arg


parser = argparse.ArgumentParser()
parser.add_argument('--project', '-p', type=_arg_directory, default=None, help='Path to project.')

req_grp = parser.add_argument_group(title='Required')
req_grp.add_argument('--pipeline_config', type=_arg_pipeline_config, help='Pipeline config .yml file.', required=True)

args = parser.parse_args()

Run(pipeline_config=args.pipeline_config, project_dir=args.project)



DAG_NAME_KEYNAME = 'PIPELINE_NAME'
PROJECT_DIR = ''
PROJECT_DIR_KEYNAME = 'PROJECT_DIR'

# STEPS INSTRUCTIONS
STEPS_KEYNAME = 'steps'
DEPENDS_ON_KEYNAME = 'depends_on'
EXEC_KEYNAME = 'exec'
EXEC_PATH_SEPARATOR = ':'



























# # SETUP INSTRUCTIONS

# SETUP_KEYNAME = 'setup'


# DAGESTATOR_SETUP_KEYNAME = 'dagestator'

# DAGESTATOR_STRICT_MODE_KEYNAME = 'STRICT_MODE'
# DAGESTATOR_STRICT_MODE_DEFAULT = False

# DISABLE_LOW_EVENTS_KEYNAME = 'ENABLE_LOW_EVENTS'
# DISABLE_LOW_EVENTS_DEFAULT = False

# DAGESTATOR_LOG_DIR_KEYNAME = 'LOG_DIR'
# DAGESTATOR_LOG_DIR_DEFAULT = None

# DAGESTATOR_LOG_FILENAME_KEYNAME = 'LOG_FILENAME'
# DAGESTATOR_LOG_FILENAME_DEFAULT = None

# DAGESTATOR_LOG_LEVEL_KEYNAME = 'LOG_LEVEL'
# DAGESTATOR_LOG_LEVEL_DEFAULT = 'INFO'


# RUNNER_SETUP_KEYNAME = 'runner'

# PROJECT_NAME_KEYNAME = 'PROJECT_NAME'

# ENABLE_MP_KEYNAME = 'ENABLE_MP'
# ENABLE_MP_DEFAULT = False

# ITER_SLEEP_SEC_KEYNAME = 'ITER_SLEEP_SEC'
# ITER_SLEEP_SEC_DEFAULT = 1

# FREE_CPU_COUNT_KEYNAME = 'FREE_CPU_COUNT'
# FREE_CPU_COUNT_DEFAULT = 1

# RUNNER_STRICT_MODE_KEYNAME = 'STRICT_MODE'
# RUNNER_STRICT_MODE_DEFAULT = False

# DAG_NAME_KEYNAME = 'PIPELINE_NAME'
# DAG_NAME_DEFAULT = 'dag'

# DAG_CONFIG_KEYNAME = 'DAG_CONFIG'
# DAG_CONFIG_DIR_KEYNAME = 'DAG_CONFIG_DIR'

# RUNNER_LOG_DIR_KEYNAME = 'LOG_DIR'
# RUNNER_LOG_DIR_DEFAULT = 'wombat_log'

# RUNNER_LOG_FILENAME_KEYNAME = 'LOG_FILENAME'
# RUNNER_LOG_FILENAME_DEFAULT = 'runner.log'

# RUNNER_LOG_LEVEL_KEYNAME = 'LOG_LEVEL'
# RUNNER_LOG_LEVEL_DEFAULT = 'INFO'

# LOCAL_STORAGE_DIR_KEYNAME = 'LOCAL_STORAGE_DIR'
# LOCAL_STORAGE_DIR_DEFAULT = 'wombat_storage'

# MAX_DISK_SPACE_USAGE_BYTES_KEYNAME = 'MAX_DISK_SPACE_USAGE_BYTES'
# MAX_DISK_SPACE_USAGE_PERCENT_KEYNAME = 'MAX_DISK_SPACE_USAGE_PERCENT'

# UPLOAD_CHUNK_BYTES_KEYNAME = 'UPLOAD_CHUNK_BYTES'
# UPLOAD_CHUNK_BYTES_DEFAULT = 100 * 1024 * 1024  # 100 MB

# HASH_CHUNK_BYTES_KEYNAME = 'HASH_CHUNK_BYTES'

# SENTRY_DNS_KEYNAME = 'SENTRY_DNS'

# S3_BUCKET_NAME_KEYNAME = 'S3_BUCKET_NAME'
# S3_BUCKET_NAME_DEFAULT = 'test-s3-bucket'

# S3_ENDPOINT_URL_KEYNAME = 'S3_ENDPOINT_URL'
# S3_ENDPOINT_URL_DEFAULT = 'https://storage.yandexcloud.net'

# RUN_TIMESTAMP_KEYNAME = 'RUN_TIMESTAMP'
# S3_ACCESS_KEY_ID_KEYNAME = 'S3_ACCESS_KEY_ID'
# S3_ACCESS_KEY_KEYNAME = 'S3_ACCESS_KEY'
# S3_CREDENTIALS_PATH_KEYNAME = 'AWS_SHARED_CREDENTIALS_FILE'
# S3_CONFIG_PATH_KEYNAME = 'AWS_CONFIG_FILE'
# S3_MAX_ATTEMPTS_KEYNAME = 'S3_MAX_ATTEMPTS'
# S3_MAX_ATTEMPTS_DEFAULT = 10

# # CACHE OPTIONS
# CACHE_OPTIONS_KEYNAME = 'CACHE'
# CACHE_STORING_KEYNAME = 'STORE'
# CACHE_STORING_DEFAULT = True
# CACHE_USE_KEYNAME = 'USE'
# CACHE_USE_DEFAULT = False

# CACHE_INVALIDATE_KEYNAME = 'INVALIDATE_BY'
# CACHE_INVALIDATE_BY_TIME_KEYNAME = 'TIME'
# CACHE_INVALIDATE_BY_TIME_DEFAULT = None
# CACHE_INVALIDATE_BY_UNCOMPLITED_KEYNAME = 'SUCCESS'
# CACHE_INVALIDATE_BY_UNCOMPLITED_DEFAULT = True
# CACHE_INVALIDATE_BY_USED_KEYNAME = 'USED_COUNT'
# CACHE_INVALIDATE_BY_BOUND_KEYNAME = 'BOUND'
# CACHE_INVALIDATE_BY_RUNTIME_KEYNAME = 'RUNTIME'
# CACHE_INVALIDATE_BY_ON_BUILD_KEYNAME = 'ON_BUILD'
# CACHE_INVALIDATE_BY_ON_ARTS_KEYNAME = 'ON_ARTS'
# CACHE_INVALIDATE_BY_ON_RUN_KEYNAME = 'ON_RUN'

# # ON ERROR
# ON_ERROR_KEYNAME = 'ON_ERROR'
# ON_ERROR_EXEC_KEYNAME = 'EXEC'
# ON_ERROR_EXEC_DEFAULT = None
# ON_ERROR_REPEAT_KEYNAME = 'REPEAT'
# ON_ERROR_REPEAT_DEFAULT = 0
# ON_ERROR_HAVE_ATTEMPTS_KEYNAME = 'attempts_count'
# ON_ERROR_PAUSE_KEYNAME = 'PAUSE'
# ON_ERROR_USED_PAUSE_KEYNAME = 'USED_PAUSE'
# ON_ERROR_PAUSE_DEFAULT = '0s'
# ON_ERROR_TIME_ERROR_KEYNAME = 'last_time'
# ON_ERROR_TOTAL_KEYNAME = 'total'
# ON_ERROR_BY_EXEC_KEYNAME = 'by_exec'


# # CONTEXT INSTRUCTIONS
# CONTEXT_KEYNAME = 'context'

# CUBE_NAME_KEYNAME = 'CUBE_NAME'  # FIXME: Кажется это внутренний ключ, надо бы их как-то иначе обозначать

# CONTEXT_GENERATORS = 'context_generator'

# ARTEFACT_INFO_REPORT_COUNT_KEYNAME = 'ARTEFACT_INFO_REPORT_COUNT'
# ARTEFACT_INFO_REPORT_COUNT_DEFAULT = 30

# COLLECT_MEM_USAGE_KEYNAME = 'COLLECT_MEM_USAGE'
# COLLECT_MEM_USAGE_DEFAULT = False


# # STEPS INSTRUCTIONS
# STEPS_KEYNAME = 'steps'
# CUBES_KEYNAME = 'cubes'
# DEPENDS_ON_KEYNAME = 'depends_on'
# TAGS_KEYNAME = 'tags'

# EXEC_KEYNAME = 'exec'
# PARAMS_KEYNAME = 'context'
# GENERATE_WITH_KEYNAME = 'generate_with'

# INDEXED_INPUTS_KEYNAME = 'inputs'
# NAMED_INPUTS_KEYNAME = 'named_inputs'
# ALL_INPUTS_KEYNAME = '*'


# # OTHER CONST

# ONECUBE_AUTO_NAME = '__cube__'
# ALL_OUTPUTS_INDEX = f'{ALL_INPUTS_KEYNAME}_indexed'
# ALL_OUTPUTS_NAMED_INDEX = f'{ALL_INPUTS_KEYNAME}_named'

# EXEC_PATH_SEPARATOR = ':'
# TAG_EXP_START = '@'

# DAG_SEPARATOR = ':'
# CUBE_SEPARATOR = '.'
# INDEX_SEPARATOR = '['
# INDEX_FINISHER = ']'
# LIST_SEPARATOR = ','
# OUTPUT_OPTION_SEPARATOR = '^'
# OUTPUT_OPTION_NOT_ONLY_SUCCESS = 'not_only_success'
# OUTPUT_OPTION_NOT_ONLY_SUCCESS_DEFAULT = False
# OUTPUT_OPTION_MIN_DATE = 'min_date'
# OUTPUT_OPTION_MIN_DATE_DEFAULT = None
# OUTPUT_OPTION_MAX_DATE = 'max_date'
# OUTPUT_OPTION_MAX_DATE_DEFAULT = None
# OUTPUT_OPTION_COUNT_RUNS = 'count_runs'
# OUTPUT_OPTION_COUNT_RUNS_DEFAULT = -1
# OUTPUT_OPTION_COUNT_RUNS_INF = '*'

# INTERSECT_SET_OPERATOR = '&'
# UNITE_SET_OPERATOR = '|'
# DIFF_SET_OPERATOR = '-'
# SYM_DIFF_SET_OPERATOR = '^'
# # NOT_SET_OPERATOR = '^'



from collections import Mapping

def deep_update(source, overrides):
    """Update a nested dictionary or similar mapping. Modify `source` in place."""
    for key, value in overrides.items():
        if isinstance(value, Mapping) and value:
            # if key not in source:
            #     source[key] = {}
            # s_c = source[key]
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
    return source


from schema import Optional, Or, Regex, Schema, Use
from ppline.utils.const import dagmap_consts

def regex_safe(text: str):
    for symbol in ('.', '[', ']', '|', '^'):
        text = text.replace(symbol, f'\{symbol}')
    return rf'{text}'


def complete(regex):
    return rf'^{regex}\Z'


exec_path_separator = regex_safe(dagmap_consts.EXEC_PATH_SEPARATOR)


STAGE_NAMING_REGEX = r'[a-zA-Z_][-\w]*'
EXEC_REGEX = rf'\S*{exec_path_separator}[a-zA-Z_]\w*'

STEP_NAME = Regex(complete(STAGE_NAMING_REGEX), error='Invalid step name.')
EXEC = Regex(complete(EXEC_REGEX), error='Exec path should be patterned as "{module_path.py}:{exec_name}".')


STATIC_DAG_SCHEMA = Schema({
    dagmap_consts.STEPS_KEYNAME: {
            STEP_NAME: {
                dagmap_consts.EXEC_KEYNAME: EXEC,
                # Optional('params'): CONTEXT_SCHEMA,
                # Optional(dagmap_consts.INDEXED_INPUTS_KEYNAME): Or([OUTPUT], [INDEXED_OUTPUT], []),

                # Optional(dagmap_consts.NAMED_INPUTS_KEYNAME): Or({
                #     dagmap_consts.ALL_INPUTS_KEYNAME: [OUTPUT]
                # }, {
                #     OUTPUT_NAME: INDEXED_OUTPUT
                # }, {})
                Optional(dagmap_consts.DEPENDS_ON_KEYNAME): Or([STEP_NAME], []),
            }
        # },
        # Optional(dagmap_consts.DEPENDS_ON_KEYNAME): Or([STEP_NAME], []),

        # Optional(dagmap_consts.TAGS_KEYNAME): Or([TAG_NAME], [])
    }
})


def parse_schema(dag: dict, schema: Schema) -> dict:
    return schema.validate(dag)



import yaml
from ppline.yamlread.validation.schemes.v1_0 import parse_schema
from ppline.yamlread.validation.schemes.v1_0 import STATIC_DAG_SCHEMA
from ppline.utils import deep_update
from ppline.utils.const import dagmap_consts



class Yamlread(object):
    def __init__(self, dag_path: str):
        self.pipeline = {}
        with open(dag_path, 'r') as inf:
            stream = '\n'.join(inf.readlines())
        try:
            deep_update(self.pipeline, yaml.load(stream, yaml.Loader))
        except yaml.YAMLError:
            raise Exception("Invalid YAML.")
    
    def parse_pipeline(self) -> dict:
        """Parse a raw user pipeline. Currently just validate the schema.

        :param dict pipeline: A raw pipeline passed to Dagestator.
        :return dict: The parsed pipeline.
        """
        self.pipeline= parse_schema(self.pipeline, STATIC_DAG_SCHEMA)

    def extract_executables(self):
        
        self.commands = [val[dagmap_consts.EXEC_KEYNAME] for val in self.pipeline[dagmap_consts.STEPS_KEYNAME].values()]

    def executor(self):
        for i in range(len(self.commands)):
            _file=self.commands[i].split(':')[0][:-3].replace('/', '.')
            _class = self.commands[i].split(':')[1]
            exec(f'from {_file} import {_class}')
            
            exec(f'c{i} = {_class}()')
            exec(f'c{i}()')

    def __call__(self):
        self.parse_pipeline()
        self.extract_executables()
        self.executor()



        






import setuptools

PACKAGE_NAME='ppline'
PACKAGE_AUTHOR='5x12'
PACKAGE_AUTHOR_EMAIL='contact@awolf.io'
PACKAGE_DESCR='Pipeline framework.'

try:
   from ppline.version import __version__ as version
except ImportError:
   exec(f'from {PACKAGE_NAME}.version import __version__ as version')

def parse_requirements(filename):
    lines = (line.strip() for line in open(filename))
    return [line for line in lines if line and not line.startswith("#")]

with open('README.md', 'r') as f: long_description=f.read()

setuptools.setup(
   name=PACKAGE_NAME,
   version=version,
   author=PACKAGE_AUTHOR,
   url = 'https://github.com/5x12/ppline',
   download_url = f'https://github.com/5x12/ppline/archive/{version}.tar.gz',
   description=PACKAGE_DESCR,
   long_description=long_description,
   license='MIT',
   long_description_content_type='text/markdown',
   packages=setuptools.find_packages(),
   install_requires=parse_requirements('requirements.txt'),
   keywords = ['PIPELINE', 'PIPE'],
   python_requires='>=3.6',
   include_package_data=True,
   classifiers=[
    'Development Status :: 3 - Alpha',   
    'Intended Audience :: Developers',      
    'License :: OSI Approved :: MIT License',   
    'Programming Language :: Python :: 3.6',
  ],
)



from typing import Optional
import os
from ppline.yamlread.yamlread import Yamlread

class Run:
    def __init__(self, pipeline_config=str, project_dir: Optional[str] = None):
        
        self.pipeline_config = pipeline_config
        self.project_dir = project_dir
        
        if self.project_dir is not None:
            path = self.project_dir+'/'+self.pipeline_config
            if os.path.exists(path)==False:
                raise Exception(f'Cannot find a config file at path {path}')
            PROJECT_DIR_KEYNAME=path
            return PROJECT_DIR_KEYNAME
        else:
            PROJECT_DIR_KEYNAME = self.pipeline_config
            
        execute=Yamlread(dag_path=PROJECT_DIR_KEYNAME)
        execute()